@@ -8,22 +8,18 @@ import functools
 import re
 import sys
 import textwrap
+import warnings
 from operator import attrgetter
 from typing import FrozenSet, Match, Tuple
 
 import autoflake
 import black
 import isort
-import libcst
 import pyupgrade._main
 from black.mode import TargetVersion
 from black.parsing import lib2to3_parse
-from pybetter.cli import (
-    ALL_IMPROVEMENTS,
-    FixMissingAllAttribute,
-    FixParenthesesInReturn,
-    FixTrivialNestedWiths,
-)
+
+from ._codemods import _run_codemods  # type: ignore
 
 
 def _fallback(source: str, **kw: object) -> Tuple[str, object]:
@@ -36,14 +32,6 @@ except ImportError:  # pragma: no cover  # on Python 3.9
     assert sys.version_info < (3, 9)
     _teyit_refactor = _fallback
 
-try:
-    from hypothesis.extra.codemods import refactor as _hypothesis_refactor
-except ImportError:  # pragma: no cover  # optional integration
-
-    def _hypothesis_refactor(source_code: str) -> str:
-        return source_code
-
-
 # We can't use a try-except here because com2ann does not declare python_requires,
 # and so it is entirely possible to install it on a Python version that it does
 # not support, and nothing goes wrong until you call the function.  We therefore
@@ -53,7 +41,7 @@ if sys.version_info[:2] >= (3, 8):  # pragma: no cover
     from com2ann import com2ann
 
 
-__version__ = "0.3.8"
+__version__ = "0.5.1"
 __all__ = ["shed", "docshed"]
 
 _version_map = {
@@ -61,19 +49,27 @@ _version_map = {
     for k in TargetVersion
     if k.value >= TargetVersion.PY36.value
 }
-_pybetter_fixers = tuple(
-    fix().improve
-    for fix in set(ALL_IMPROVEMENTS)
-    - {FixMissingAllAttribute, FixParenthesesInReturn, FixTrivialNestedWiths}
+_default_min_version = min(_version_map.values())
+_SUGGESTIONS = (
+    # If we fail on invalid syntax, check for detectable wrong-codeblock types
+    (r"^(>>> | In [\d+]: )", "pycon"),
+    (r"^Traceback \(most recent call last\):$", "python-traceback"),
 )
 
 
+class ShedSyntaxWarning(SyntaxWarning):
+    """Warns that shed has been called on something with invalid syntax."""
+
+
 @functools.lru_cache()
 def shed(
     source_code: str,
     *,
     refactor: bool = False,
     first_party_imports: FrozenSet[str] = frozenset(),
+    min_version: Tuple[int, int] = _default_min_version,
+    _location: str = "string passed to shed.shed()",
+    _remove_unused_imports: bool = True,
 ) -> str:
     """Process the source code of a single module."""
     assert isinstance(source_code, str)
@@ -81,20 +77,48 @@ def shed(
     assert isinstance(first_party_imports, frozenset)
     assert all(isinstance(name, str) for name in first_party_imports)
     assert all(name.isidentifier() for name in first_party_imports)
+    assert min_version in _version_map.values()
 
     if source_code == "":
         return ""
 
+    format_docs = functools.partial(
+        docshed,
+        refactor=refactor,
+        first_party_imports=first_party_imports,
+        min_version=min_version,
+        _location=_location,
+    )
+
     # Use black to autodetect our target versions
-    target_versions = {
-        v
-        for v in black.detect_target_versions(
-            lib2to3_parse(source_code.lstrip(), set(_version_map))
+    try:
+        parsed = lib2to3_parse(source_code.lstrip(), set(_version_map))
+        # black.InvalidInput, blib2to3.pgen2.tokenize.TokenError, SyntaxError...
+        # for forwards-compatibility I'm just going general here.
+    except Exception:
+        msg = f"Could not parse {_location}"
+        for pattern, blocktype in _SUGGESTIONS:
+            if re.search(pattern, source_code, flags=re.MULTILINE):
+                msg += f"\n    Perhaps you should use a {blocktype!r} block instead?"
+        try:
+            compile(source_code, "<string>", "exec")
+        except SyntaxError:
+            pass
+        else:
+            msg += "\n    The syntax is valid Python, so please report this as a bug."
+        warnings.warn(
+            ShedSyntaxWarning(msg),
+            stacklevel=_location.count(" block in ") + 2,
         )
-        if v.value >= TargetVersion.PY36.value
-    }
+        # Even if the code itself has invalid syntax, we might be able to
+        # regex-match and therefore reformat code embedded in docstrings.
+        return format_docs(source_code)
+    target_versions = set(_version_map) & set(black.detect_target_versions(parsed))
     assert target_versions
-    min_version = _version_map[min(target_versions, key=attrgetter("value"))]
+    min_version = max(
+        min_version,
+        _version_map[min(target_versions, key=attrgetter("value"))],
+    )
 
     if refactor:
         # Some tools assume that the file is multi-line, but empty files are valid input.
@@ -108,23 +132,8 @@ def shed(
         )
         # Use teyit to replace old unittest.assertX methods on Python 3.9+
         source_code, _ = _teyit_refactor(source_code)
-        # Apply Hypothesis codemods to fix any deprecated code
-        source_code = _hypothesis_refactor(source_code)
-        # Then apply pybetter's fixes with libcst
-        tree = libcst.parse_module(source_code)
-        for fixer in _pybetter_fixers:
-            try:
-                # Might raise e.g. https://github.com/Instagram/LibCST/issues/446
-                newtree = fixer(tree)
-            except Exception:
-                pass
-            else:
-                tree = newtree
-                # Catches e.g. https://github.com/lensvol/pybetter/issues/60
-                compile(newtree.code, "<string>", "exec")
-        source_code = tree.code
-    # Then shed.docshed (below) formats any code blocks in documentation
-    source_code = docshed(source=source_code, first_party_imports=first_party_imports)
+    # Apply all our libcst-based codemods
+    source_code = _run_codemods(source_code, refactor=refactor)
     # And pyupgrade - see pyupgrade._main._fix_file - is our last stable fixer
     # Calculate separate minver because pyupgrade can take a little while to update
     pyupgrade_min = min(min_version, max(pyupgrade._main.IMPORT_REMOVALS))
@@ -145,9 +154,7 @@ def shed(
         source_code = autoflake.fix_code(
             source_code,
             expand_star_imports=True,
-            remove_all_unused_imports=True,
-            remove_duplicate_keys=True,
-            remove_unused_variables=True,
+            remove_all_unused_imports=_remove_unused_imports,
         )
         source_code = isort.code(
             source_code,
@@ -157,6 +164,8 @@ def shed(
             combine_as_imports=True,
         )
 
+    # Then shed.docshed (below) formats any code blocks in documentation
+    source_code = format_docs(source_code)
     # Remove any extra trailing whitespace
     return source_code.rstrip() + "\n"
 
@@ -167,6 +176,8 @@ def docshed(
     *,
     refactor: bool = False,
     first_party_imports: FrozenSet[str] = frozenset(),
+    min_version: Tuple[int, int] = _default_min_version,
+    _location: str = "string passed to shed.docshed()",
 ) -> str:
     """Process Python code blocks embedded in documentation."""
     # Inspired by the blacken-docs package.
@@ -174,6 +185,15 @@ def docshed(
     assert isinstance(first_party_imports, frozenset)
     assert all(isinstance(name, str) for name in first_party_imports)
     assert all(name.isidentifier() for name in first_party_imports)
+    assert min_version in _version_map.values()
+    format_code = functools.partial(
+        shed,
+        refactor=refactor,
+        first_party_imports=first_party_imports,
+        min_version=min_version,
+        _remove_unused_imports=False,
+    )
+
     markdown_pattern = re.compile(
         r"(?P<before>^(?P<indent> *)```python\n)"
         r"(?P<code>.*?)"
@@ -182,8 +202,10 @@ def docshed(
     )
     rst_pattern = re.compile(
         r"(?P<before>"
-        r"^(?P<indent> *)\.\. (jupyter-execute::|(code|code-block|sourcecode|ipython):: "
-        r"(python|py|sage|python3|py3|numpy))\n"
+        r"^(?P<indent> *)\.\. "
+        r"(?P<block>jupyter-execute::|"
+        r"invisible-code-block: python|"  # magic rst comment for Sybil doctests
+        r"(code|code-block|sourcecode|ipython):: (python|py|sage|python3|py3|numpy))\n"
         r"((?P=indent) +:.*\n)*"
         r"\n*"
         r")"
@@ -193,7 +215,7 @@ def docshed(
 
     def _md_match(match: Match[str]) -> str:
         code = textwrap.dedent(match["code"])
-        code = shed(code, refactor=refactor, first_party_imports=first_party_imports)
+        code = format_code(code, _location=f"```python markdown block in {_location}")
         code = textwrap.indent(code, match["indent"])
         return f'{match["before"]}{code}{match["after"]}'
 
@@ -205,7 +227,9 @@ def docshed(
         assert trailing_ws_match
         trailing_ws = trailing_ws_match.group()
         code = textwrap.dedent(match["code"])
-        code = shed(code, refactor=refactor, first_party_imports=first_party_imports)
+        code = format_code(
+            code, _location=f"{match['block']!r} rst block in {_location}"
+        )
         code = textwrap.indent(code, min_indent)
         return f'{match["before"]}{code.rstrip()}{trailing_ws}'
 
