@@ -9,12 +9,13 @@ import functools
 import multiprocessing
 import subprocess
 import sys
+import warnings
 from pathlib import Path
 from typing import FrozenSet, Union
 
 import autoflake
 
-from . import docshed, shed
+from . import ShedSyntaxWarning, _version_map, docshed, shed
 
 
 @functools.lru_cache()
@@ -51,17 +52,26 @@ def _rewrite_on_disk(
         # Permissions or encoding issue, or file deleted since last commit.
         return f"skipping {fname!r} due to {err}"
     writer = docshed if fname.endswith((".md", ".rst")) else shed
+    msg = ""
     try:
-        result = writer(on_disk, **kwargs)
+        with warnings.catch_warnings(record=True) as record:
+            result = writer(on_disk, _location=f"file {fname!r}", **kwargs)
     except Exception as err:  # pragma: no cover  # bugs are unknown xor fixed ;-)
         return (
-            f"Internal error formatting {fname!r}: {err}\n"
+            f"Internal error formatting {fname!r}: {type(err).__name__}: {err}\n"
             "    Please report this to https://github.com/Zac-HD/shed/issues"
         )
+    else:
+        msg = "\n".join(
+            str(w.message) for w in record if issubclass(w.category, ShedSyntaxWarning)
+        )
+        for w in record:  # pragma: no cover  # just being careful not to hide anything
+            if not issubclass(w.category, ShedSyntaxWarning):
+                warnings.warn(w.message, category=w.category, stacklevel=2)
     if result != on_disk:
         with open(fname, mode="w") as fh:
             fh.write(result)
-    return result != on_disk
+    return msg or result != on_disk
 
 
 def cli() -> None:  # pragma: no cover  # mutates things in-place, will test later.
@@ -79,6 +89,16 @@ def cli() -> None:  # pragma: no cover  # mutates things in-place, will test lat
         dest="files",
         help="File(s) to format, instead of autodetection",
     )
+    min_version_group = parser.add_mutually_exclusive_group()
+    oldest, *rest = sorted(_version_map.values())
+    parser.set_defaults(min_version=oldest)
+    for version in rest:
+        min_version_group.add_argument(
+            f"--py3{version[1]}-plus",
+            action="store_const",
+            dest="min_version",
+            const=version,
+        )
     args = parser.parse_args()
 
     if args.files:
@@ -93,7 +113,7 @@ def cli() -> None:  # pragma: no cover  # mutates things in-place, will test lat
                 stdout=subprocess.PIPE,
                 universal_newlines=True,
             ).stdout.splitlines()
-        except subprocess.SubprocessError:
+        except (subprocess.SubprocessError, FileNotFoundError):
             print("Doesn't seem to be a git repo; pass filenames to format.")  # noqa
             sys.exit(1)
         all_filenames = [f for f in all_filenames if _should_format(f)]
@@ -102,17 +122,26 @@ def cli() -> None:  # pragma: no cover  # mutates things in-place, will test lat
         _rewrite_on_disk,
         first_party_imports=_guess_first_party_modules(),
         refactor=args.refactor,
+        min_version=args.min_version,
     )
 
-    if len(all_filenames) <= 4:
-        # If we're only formatting a few files, starting up a process pool
-        # probably takes up more time that it saves.
-        for fname in all_filenames:
-            error_msg = rewrite(fname)
-            if isinstance(error_msg, str):
-                print(error_msg)  # noqa
-    else:
-        with multiprocessing.Pool() as pool:
-            for error_msg in pool.imap_unordered(rewrite, all_filenames):
-                if isinstance(error_msg, str):
-                    print(error_msg)  # noqa
+    if len(all_filenames) > 4:
+        # If we're formatting more than a few files, the improved throughput
+        # of a process pool probably covers the startup cost.
+        try:
+            with multiprocessing.Pool() as pool:
+                for error_msg in pool.imap_unordered(rewrite, all_filenames):
+                    if isinstance(error_msg, str):
+                        print(error_msg)  # noqa
+            return
+        except BlockingIOError as err:  # pragma: no cover
+            # This can occur when `os.fork()` fails due to limited available
+            # memory or number-of-processes.  In this case, we fall back to
+            # the slow path; and reprocess whatever we've already done for
+            # simplicity.  See https://stackoverflow.com/q/44534288/
+            print(f"Error: {err!r}\n    Falling back to serial mode.")  # noqa
+
+    for fname in all_filenames:
+        error_msg = rewrite(fname)
+        if isinstance(error_msg, str):
+            print(error_msg)  # noqa
